/* app_main.h - Defines and prototypes of functions commonly used across the application */

/*
 * Copyright (c) 2017-23 Vithamas Technologies Private Limited
 *
 */
#ifndef __APP_COMMON_H__
#define __APP_COMMON_H__

/* Includes --------------------------------------------------------------------------------------*/
#include <errno.h>
#include <zephyr/types.h>
#include <zephyr/kernel.h>
#include <zephyr/sys/printk.h>
#include <zephyr/settings/settings.h>
#include <zephyr/bluetooth/gatt.h>
#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/device.h>
#include <zephyr/sys/reboot.h>
#include <bluetooth/services/ddfs.h>
#include <dm.h>
#include <hal/nrf_gpio.h>
#include <stddef.h>
#include <zephyr/sys/util.h>
#include <zephyr/bluetooth/hci_vs.h>
#include "led_pwm.h"
#include "buzzer_pwm.h"

/* Macros ----------------------------------------------------------------------------------------*/
/* FW Versions */
#define FW_VERSION_MAJOR                (0)
#define FW_VERSION_MINOR                (1)

#define MAX_BEACONS_SUPPORTED           (4)
#define MAX_TX_POWER                    (8) // make sure the same is set in the prj.conf file
#define BATTERY_LOW_THRESHOLD           (20)

/* Feature/R&D macros */
#define TEST_DFU                        (0)
#define ENABLE_DM                       (1)

/* Range zones definition */
#define ZONE_1		                    (5)  // mts
#define ZONE_2		                    (10) // mts
#define ZONE_3		                    (15) // mts
#define ZONE_SILENT                     (25) // mts

/*Range zone definition for mobile....*/
#define ZONE1_RSSI                      (-75)
#define ZONE2_RSSI                      (-85)
#define ZONE3_RSSI                      (-90)
#define ZONE_SILENT_RSSI                (-97)
#define MAX_RSSI_SAMPLES_COUNT          (5)

#define DEVICE_NAME                     CONFIG_BT_DEVICE_NAME
#define DEVICE_NAME_LEN                 (sizeof(DEVICE_NAME) - 1)

#define MAX_BLE_DATA_SZ			        (16)

// Byte offset for the BLE data packet being exchanged
//
#define BYTE_OFF_OPCODE 		        (0)
#define BYTE_OFF_LEN    		        (1)
#define BYTE_OFF_DATA   		        (2)
#define BLE_PKT_HEADER_SZ		        (2)
#define MAX_ZONE                        (4)

/* Application timers */
#define MAX_DEVICE_TIMERS_SUPPORTED     (MAX_BEACONS_SUPPORTED + 1) // additional timer for current device
#define ZONE_CHANGE_TIMEOUT_MSEC        (5000)
#define MODE_CHANGE_TIMEOUT_MSEC        (10000)
#define ASSOCIATION_TIMEOUT_MSEC        (30000)
#define PAIRING_TIMEOUT_MSEC            (30000)
#define DM_TIME_INTERVAL_MSEC           (2000)
#define BATTERY_MONITOR_INTERVAL_SECS   (600)

/* Global param ----------------------------------------------------------------------------------*/
/* OPCODES */
typedef enum {
    OPCODE_DUMMY_WRITE,		// a dummy gatt write to the CCCD to enable indication
	OPCODE_CONFIRM_PARING,
	OPCODE_START_ALERT,
	OPCODE_STOP_ALERT,
	OPCODE_SET_ZONE,
	OPCODE_DM_RANGING,
    OPCODE_DISSOCIATE,
    OPCODE_OTA_FOR_DETECTION_DEVICE,
    OPCODE_OTA_FOR_BEACON_DEVICE,
    OPCODE_FW_VER_INFO,
    OPCODE_SET_SILENT_MODE,
    OPCODE_GET_MOBILE_ID,
    OPCODE_LOW_BATTERY,
    OPCODE_FIND_DEVICE,
    OPCODE_DEVICE_FOUND,
    OPCODE_DISABLE_SILENT_MODE,
} OPCODES_T;

typedef enum
{
	DEVICE_MODE_BEACON,	  /** Peripheral device */
	DEVICE_MODE_DETECTION /** Central device */
} DEVICE_MODE_T;

typedef enum
{
	EVENT_NONE,                         // 0
	//
	// Add all the button events below
	//
	EVENT_BTN_SINGLE_PRESS,             // 1
	EVENT_BTN_DOUBLE_PRESS,             // 2
	EVENT_BTN_TRIPLE_PRESS,             // 3
	EVENT_BTN_LONG_PRESS_6SECS,         // 4
	EVENT_START_DM,                     // 5
	EVENT_DM_RESULT_READY,              // 6
    EVENT_BT_DISCONNECT,                // 7
    EVENT_UPDATE_NVS,                   // 8
    EVENT_UPDATE_STATE,                 // 9
	EVENT_PERIODIC_DM_REQ,              // 10
    EVENT_PROCESS_BLE_DATA,             // 11
    EVENT_PAIRED_DEVICE_CONNECTED,      // 12
    EVENT_DISSOCIATE,                   // 13
    EVENT_BATTERY_MONITOR,              // 14
} APP_TASK_EVENTS_T;

typedef enum
{
	APP_STATE_FACTORY_RESET,        // 0
	APP_STATE_IDLE,                 // 1
	APP_STATE_MODE_TRANSIT,         // 2
	APP_STATE_PAIRING,              // 3
    APP_STATE_GET_DEVICE_ID,        // 4
    APP_STATE_CHECK_EXISTENCE,      // 5
	APP_STATE_WAIT_CONFIRM,         // 6

	// !!! DM ranging starts after pair successful and below states supports DM ranging. !!!
	//
	APP_STATE_PAIRED,               // 7
	APP_STATE_ZONE_RANGING,         // 8
	APP_STATE_ZONE_SETUP,           // 9
	APP_STATE_ALERT_START,          // A
	APP_STATE_ALERT_STOP,           // B
    APP_STATE_SETUP_FOR_OTA,        // C
    APP_STATE_OTA_UPDATE,           // D
    APP_STATE_SETUP_BEACON_FOR_OTA, // E
    APP_STATE_HOMING,               // F
    APP_STATE_SILENT_MODE,          // 10
    APP_STATE_DISABLE_SILENT_MODE,
    APP_STATE_FIND_THE_DEVICE,      // 11
    APP_STATE_DEVICE_FOUND,         // 12
} APP_STATE_T;

typedef enum {
    TIMER_EVENT_NONE,
    TIMER_EVENT_PAIRING,
    TIMER_EVENT_ASSOCIATION,
    TIMER_EVENT_DM_SYNC,
    TIMER_EVENT_ZONE_CHANGE,
    TIMER_EVENT_MODE_CHANGE,

    TIMER_EVENT_INVALID,
} TIMER_EVENT_T;

typedef struct
{
	APP_TASK_EVENTS_T event;
    APP_STATE_T state;
	void *data;
	struct dm_result result;
    int8_t device_index;
} app_task_msg_t;

typedef enum {
	CONN_STATUS_DISCONNECTED,
    CONN_STATUS_CONNECTING,
	CONN_STATUS_CONNECTED
} CONN_STATUS_T;

typedef struct {
	CONN_STATUS_T conn_status;
	bt_addr_le_t addr;
	struct bt_conn *conn;
	uint32_t rand_seed;
    float dm_mcpd_best;
	uint16_t cmd_chr_handle;
	uint8_t pair_status;
	uint8_t write_complete;
    uint8_t zone_breach_status;
    uint8_t fw_version_major;
    uint8_t fw_version_minor;
    uint8_t id[BT_ADDR_SIZE];
    int8_t rssi_avg[MAX_RSSI_SAMPLES_COUNT];
    uint8_t rssi_off;
    bool is_mobile;
} paired_dev_info_t;

typedef struct {
	uint8_t opcode;
	uint8_t data[MAX_BLE_DATA_SZ];
	uint8_t len;
} message_context_t;

typedef union {
    uint8_t arr[64];

	// !!! Make sure the below structure size do not exceed the arr size !!!
	//
    struct
    {
        struct {
            uint8_t id[BT_ADDR_SIZE];
            bool is_mobile;
        } pair_info[MAX_BEACONS_SUPPORTED];

        DEVICE_MODE_T mode;             // configured device mode (central/peripheral)
        uint8_t zone_range;             // configured zone range
        uint8_t paired_device_count;    // total number of devices paired with
    } param;
} app_nvs_t;

typedef struct {
    uint8_t zone_range;
    uint8_t zone_range_far;
    uint8_t zone_range_mid;
    uint8_t zone_range_near;
    APP_STATE_T state;
    uint16_t conn_interval;
    uint16_t zone_change_count;
    uint8_t alert_status;
    uint8_t silent_mode;
    uint8_t ack;
    int8_t rssi;
    int8_t rssi_range;
    bool is_advertising;
} device_information_t;

typedef void (*timeout_callback_func)(struct k_timer *p_timer);
typedef struct {
    bool is_active;
    TIMER_EVENT_T event;
    timeout_callback_func timeout_cb;
    k_timeout_t timeout_ms;
    struct k_timer *p_timer;
} device_timers_t;

typedef enum {
    CURRENT_DEVICE,
    PERIPHERAL_1,
    PERIPHERAL_2,
    PERIPHERAL_3,
    PERIPHERAL_4,
    PERIPHERAL_5,

    TIMER_ID_INVALID,
} timer_id_t;

/* Extern param ----------------------------------------------------------------------------------*/
extern struct k_msgq app_msgq;
extern paired_dev_info_t paired_list[MAX_BEACONS_SUPPORTED];
extern paired_dev_info_t current_dev_info;
extern int8_t g_current_index;
extern app_nvs_t app_nvs;
extern message_context_t msg_ctx;
extern device_information_t dev_info;
extern bt_addr_le_t *gp_dissociate_addr;
extern volatile bool gb_scanning_in_progress;
extern volatile bool gb_dm_write_resp_wait;

/* Timers */
extern device_timers_t timer_ctx[MAX_DEVICE_TIMERS_SUPPORTED];
extern timer_id_t timer_id;
extern struct k_timer led_timer;
extern struct k_timer dm_timer;

/* Global function prototypes --------------------------------------------------------------------*/
/* Timer callback functions prototype */
void cb_battery_monitor_timer (struct k_timer *p_timer);
void cb_conn_interval_timer (struct k_timer *p_timer);
void cb_current_device_timer (struct k_timer *p_timer);
void cb_peripheral1_timer (struct k_timer *p_timer);
void cb_peripheral2_timer (struct k_timer *p_timer);
void cb_peripheral3_timer (struct k_timer *p_timer);
void cb_peripheral4_timer (struct k_timer *p_timer);
void app_update_state (APP_STATE_T new_state);
void app_thread (void);
void discover_services (void);
void write_resp_cb (struct bt_conn *conn, uint8_t err, struct bt_gatt_write_params *params);
void app_gatt_write (int8_t dev_index, uint8_t *data, uint16_t len);
uint8_t indication_cb (struct bt_conn *conn, struct bt_gatt_subscribe_params *params,
			           const void *data, uint16_t length);
int start_scan (void);
int stop_scan (void);
void central_send_msg_to_all_paired_devices (message_context_t *p_msg_ctx);
void init_central (void);
void set_tx_power (uint16_t handle, int8_t tx_pwr_lvl);
void read_conn_tx_power (uint16_t handle, int8_t *p_tx_pwr);
void read_conn_rssi (uint16_t handle, int8_t *rssi);
int8_t get_paired_device_slot_from_id_param (const uint8_t *id);
bool remove_paired_dev (uint8_t index);
int8_t get_paired_device_slot_from_conn_param (struct bt_conn *p_conn);
extern void detection_dm_add_request(uint32_t seed, bt_addr_le_t addr);
extern void beacon_dm_add_request (uint32_t seed, bt_addr_le_t addr);

#endif //__APP_COMMON_H__
