/*
 * Copyright (c) vithamastech
 *
 *
 */

/* Includes --------------------------------------------------------------------------------------*/
#include <zephyr/devicetree.h>
#include <zephyr/drivers/gpio.h>
#include <zephyr/drivers/adc.h>
#include <hal/nrf_gpio.h>
#include "batt_adc.h"

/* Private macros --------------------------------------------------------------------------------*/
#define ADC_NODE DT_NODELABEL   (adc)
#define REF_VOLTAGE             (0.6)
#define FULL_SCALE_READING      (4095) // ((2^12)-1) bitrate
#define ADC_RESOLUTION          (12)
#define ADC_CHANNEL             (0)
#define ADC_PORT                SAADC_CH_PSELP_PSELP_AnalogInput0
#define ADC_REFERENCE           ADC_REF_INTERNAL
#define ADC_GAIN                ADC_GAIN_1_6
#define BATT_MIN_mv		        (2300)
#define BATT_MAX_mv		        (3000)
#define TRANSISTOR_THRESHOLD_mv (0) // 0.7v

/* Private param ---------------------------------------------------------------------------------*/
static const struct device *adc_dev = DEVICE_DT_GET(ADC_NODE);
struct adc_channel_cfg ch0_cfg =
{
	.gain = ADC_GAIN,
	.reference = ADC_REFERENCE,
	.acquisition_time = ADC_ACQ_TIME(ADC_ACQ_TIME_MICROSECONDS,10),
	.channel_id = ADC_CHANNEL,
#ifdef CONFIG_ADC_NRFX_SAADC
	.input_positive = ADC_PORT,
#endif
};

/* Global param ----------------------------------------------------------------------------------*/
uint16_t sample_buffer[1];
struct adc_sequence sequence = {
	.channels = BIT(ADC_CHANNEL),
	.buffer = sample_buffer,
	.buffer_size = sizeof(sample_buffer),
	.resolution = ADC_RESOLUTION,
};

/**
 * @brief Configure ADC for battery voltage reading
 *
 */
void batt_init (void)
{
    if (device_is_ready(adc_dev)) {
        adc_channel_setup(adc_dev, &ch0_cfg);
    }
}

/**
 * @brief Calculate battery voltage using ADC RAW values and convert the current battery voltage to
 * percentage
 *
 * @return uint32_t
 */
uint8_t batt_read (void)
{
	uint8_t err;

	err = adc_read(adc_dev, &sequence);
	if (err != 0) {
		return err;
	}

	uint32_t adc_val = sample_buffer[0];
	uint32_t adc_vref = adc_ref_internal(adc_dev);
	adc_raw_to_millivolts(adc_vref, ADC_GAIN, ADC_RESOLUTION, &adc_val);

	// As per the schematic there is voltage divider with both R1 and R2 being 1M ohms at
	// ADC battery reading so, we need to multiply by 2 for getting the actual battery voltage
    //
	uint32_t adc_mv = (adc_val << 1) + TRANSISTOR_THRESHOLD_mv;

	float batt_volt_diff    = (adc_mv - BATT_MIN_mv);
	float batt_max_min_diff = (BATT_MAX_mv - BATT_MIN_mv);
	float batt_value  		= (batt_volt_diff / batt_max_min_diff) * 100;

    return (batt_value > 100) ? 100 : batt_value;
}
