/**
 * @file ble_services_peripheral.c
 * @author
 * @brief BLE GATT services and characteristics definition file
 *
 * @version 0.1
 * @date 2023-02-20
 *
 * @copyright Copyright (c) 2023
 *
 */

/* Includes --------------------------------------------------------------------------------------*/
#include <zephyr/bluetooth/uuid.h>
#include "app_common.h"
#include "ble_services_peripheral.h"
#include "app_dm.h"

/* Private param ---------------------------------------------------------------------------------*/
/* Advertising data Initialization */
static const struct bt_data ad[] = {
	BT_DATA_BYTES(BT_DATA_FLAGS, (BT_LE_AD_GENERAL | BT_LE_AD_NO_BREDR)),
	BT_DATA(BT_DATA_NAME_COMPLETE, DEVICE_NAME, DEVICE_NAME_LEN),
    BT_DATA_BYTES(BT_DATA_UUID128_ALL,
		      0x84, 0xaa, 0x60, 0x74, 0x52, 0x8a, 0x8b, 0x86,
		      0xd3, 0x4c, 0xb7, 0x1d, 0x1d, 0xdc, 0x53, 0x8d),
};

/* Scan response Data Initialization */
static const struct bt_data sd[] = {
	BT_DATA_BYTES(BT_DATA_UUID128_ALL, BT_UUID_NOTIFY_DEV_VAL),
};

struct bt_le_ext_adv *adv;
bt_addr_le_t connected_bt_dev;
bool ind_enabled;
static struct bt_gatt_indicate_params ind_params;

/* Global param ----------------------------------------------------------------------------------*/
struct bt_le_adv_param adv_param_conn =
	BT_LE_ADV_PARAM_INIT(BT_LE_ADV_OPT_CONNECTABLE|
						 BT_LE_ADV_OPT_CODED,
			             BT_GAP_ADV_FAST_INT_MIN_2,
			             BT_GAP_ADV_FAST_INT_MAX_2,
			             NULL);

struct bt_le_adv_param *adv_param = &adv_param_conn;
uint8_t g_data_buf[20];

/* Private function prototype --------------------------------------------------------------------*/
static ssize_t callback_write (struct bt_conn *, const struct bt_gatt_attr *, const void *,
                               uint16_t, uint16_t, uint8_t);
static void detection_dev_connected (struct bt_conn *conn, uint8_t conn_err);
static void detection_dev_disconnected (struct bt_conn *conn, uint8_t reason);
static void peripheral_le_param_updated(struct bt_conn *conn, uint16_t interval,
				                        uint16_t latency, uint16_t timeout);
static void notif_ccc_cfg_changed(const struct bt_gatt_attr *attr, uint16_t value);

//
// Primary Service Declaration
//
BT_GATT_SERVICE_DEFINE(notif_dev_svc,
    BT_GATT_PRIMARY_SERVICE(BT_UUID_NOTIFY_DEV),
    BT_GATT_CHARACTERISTIC(BT_UUID_CMD,
                           (BT_GATT_CHRC_WRITE | BT_GATT_CHRC_INDICATE | BT_GATT_CCC_INDICATE),
                           BT_GATT_PERM_WRITE,
                           NULL, callback_write,
                           NULL),
	BT_GATT_CCC(notif_ccc_cfg_changed, BT_GATT_PERM_READ | BT_GATT_PERM_WRITE),

);

BT_CONN_CB_DEFINE(conn_callbacks) = {
	.connected = detection_dev_connected,
	.disconnected = detection_dev_disconnected,
	.le_param_updated = peripheral_le_param_updated

};

/**
 * @brief callback_write
 * Callback function for write request/command from the peer, process the received data here
 *
 * @param conn   - connection ID
 * @param attr   - attribute to the data received
 * @param buf    - buffer to copy the received data
 * @param len    - data length to copy
 * @param offset - offset to the actual data to copy
 * @param flags
 * @return ssize_t
 */
static ssize_t callback_write (struct bt_conn *conn, const struct bt_gatt_attr *attr,
                               const void *buf, uint16_t len, uint16_t offset, uint8_t flags)
{
	static app_task_msg_t msg;

    len = len > sizeof(g_data_buf) ? sizeof(g_data_buf) : len;
    memcpy(g_data_buf, buf, len);
    msg.event = EVENT_PROCESS_BLE_DATA;
    msg.data  = g_data_buf;
    k_msgq_put(&app_msgq, &msg, K_NO_WAIT);

	return len;
}

/**
 * @brief central connected callback
 *
 * @param conn
 * @param conn_err
 */
static void detection_dev_connected (struct bt_conn *conn, uint8_t conn_err)
{
    static app_task_msg_t msg;
    if (app_nvs.param.mode == DEVICE_MODE_DETECTION) {
        return;
    }

    /* If user has initiated OTA, allow connection for all */
    if (dev_info.state == APP_STATE_OTA_UPDATE) {
        return;
    }

    /* If the device is already paired with any detection device, connect only if the peer is paired
    If not, disconnect */
	bt_addr_le_t *p_addr = (bt_addr_le_t *)bt_conn_get_dst(conn);
    paired_list[0].conn_status = CONN_STATUS_CONNECTING;
    paired_list[0].conn 	   = conn;
    if (paired_list[0].pair_status) {
        int8_t is_new_peer = memcmp(&paired_list[0].id, p_addr->a.val, BT_ADDR_SIZE);
        if (is_new_peer) {
           	msg.event = EVENT_BT_DISCONNECT;
        	k_msgq_put(&app_msgq, &msg, K_NO_WAIT);
            return;
        }
    }

	memcpy(&paired_list[0].addr , bt_conn_get_dst(conn), sizeof(connected_bt_dev));
}

/**
 * @brief Callback function, called when the peer device is disconnected with the local device
 *
 * @param conn
 * @param reason
 */
static void detection_dev_disconnected (struct bt_conn *conn, uint8_t reason)
{
    static app_task_msg_t msg;

	if (app_nvs.param.mode == DEVICE_MODE_BEACON) {

        paired_list[0].conn_status = CONN_STATUS_DISCONNECTED;
		paired_list[0].conn 	   = NULL;
        /* Disconnected from the peer, check if we were paired with the disconnected device
        If we were paired, go to pairing state and wait for reconnection
        Else, go to idle */
        if (paired_list[0].pair_status) {
			msg.state = (dev_info.silent_mode ? APP_STATE_PAIRING : APP_STATE_ALERT_START);
        }
        else {
            msg.state = APP_STATE_IDLE;
        }

		msg.event = EVENT_UPDATE_STATE;
        k_msgq_put(&app_msgq, &msg, K_NO_WAIT);
	}
}

/**
 * @brief Connection param update callback function
 *
 * @param conn
 * @param interval
 * @param latency
 * @param timeout
 */
static void peripheral_le_param_updated (struct bt_conn *conn, uint16_t interval,
				                         uint16_t latency, uint16_t timeout)
{
	dev_info.conn_interval = (interval * 1.25);
}

/**
 * @brief Callback function, called when peer device subscribe to our client characteristic
 *
 * @param attr
 * @param value
 */
static void notif_ccc_cfg_changed (const struct bt_gatt_attr *attr, uint16_t value)
{
	ARG_UNUSED(attr);

	ind_enabled = (value == BT_GATT_CCC_INDICATE);
}

/**
 * @brief Start beacon advertisement
 *
 * @return int
 */
void advertisement_start (void)
{
	if (!dev_info.is_advertising) {
    	int err = bt_le_adv_start(BT_LE_ADV_CONN, ad, ARRAY_SIZE(ad), sd, ARRAY_SIZE(sd));
    	if (err == 0) {
            dev_info.is_advertising = true;
    	}
	}

	return;
}

/**
 * @brief Stop BLE advertisement and update the status
 *
 */
void advertisement_stop (void)
{
    if (!dev_info.is_advertising) {
        return;
    }

	bt_le_adv_stop();
    dev_info.is_advertising = false;
}

/**
 * @brief Send the requested length data over GATT notification
 *
 * @param p_data - pointer to the data to send
 * @param len    - length of the data to send
 */
void bt_indicate (uint8_t *p_data, uint8_t len)
{
	if (ind_enabled) {
    	ind_params.attr = &notif_dev_svc.attrs[1];
    	ind_params.func = NULL;
    	ind_params.destroy = NULL;
    	ind_params.data = p_data;
    	ind_params.len = sizeof(p_data);
    	bt_gatt_indicate(NULL, &ind_params);
	}
}
