/**
 * @file buttons.c
 * @author
 * @brief Button driver implementation
 * @version 0.3
 * @date 2023-05-26
 *
 * @copyright Copyright (c) 2023
 *
 */

/* Includes --------------------------------------------------------------------------------------*/
#include "buttons.h"
#include <hal/nrf_gpio.h>

/* Private macros --------------------------------------------------------------------------------*/
#define BTN_DEBOUNCE_TIMEOUT_MS     (50)
#define SW0_NODE	                DT_ALIAS(sw0)
#define GPIO_BUTTON                 (32)

//
// For long press detection
// If the button is pressed for atleast 3sec, put the device to sleep
// If the button is pressed for >=6 secs, factory reset the device.
//
#define SLEEP_LONG_PRESS_MAX_CNT    (50)
#define RESET_LONG_PRESS_MAX_CNT    (80)

//
// Single press max timeout 600ms (TRIPLE_PRESS_WAIT * BTN_DEBOUNCE_TIMEOUT_MS)
//
#define DOUBLE_PRESS_WAIT           (10)
#define TRIPLE_PRESS_WAIT           (12)

/* Private param ---------------------------------------------------------------------------------*/
static struct {
    button_state_t state;
    volatile uint8_t long_press_wait;
    volatile uint8_t double_press_wait;
    volatile uint8_t triple_press_wait;
} g_btn_ctx;

//
// Configure button
//
static const struct gpio_dt_spec button = GPIO_DT_SPEC_GET_OR(SW0_NODE, gpios, {0});
static struct gpio_callback button_cb_data;

//
// Timer definition
//
K_TIMER_DEFINE(btn_debounce_timer, NULL, NULL);

/* Private function prototype --------------------------------------------------------------------*/
static void button_event_handler (uint8_t event);
static void btn_context_reset (void);

/* _______________________________________________________________________________________________*/

/**
 * @brief Button GPIO interrupt handler
 *
 * @param event
 */
static void button_event_handler (uint8_t event)
{
   app_task_msg_t msg;
   msg.event = event;
   msg.data = NULL;
   k_msgq_put(&app_msgq, &msg, K_NO_WAIT);
}

/**
 * @brief btn_context_reset
 * Reset button context, set the params to it's default values
 *
 */
static void btn_context_reset (void)
{
    g_btn_ctx.triple_press_wait = 0;
    g_btn_ctx.double_press_wait = 0;
    g_btn_ctx.long_press_wait   = 0;
    g_btn_ctx.state             = BTN_STATE_RELEASED;
}

/**
 * @brief btn_timer_callback
 * Timer expire callback function
 *
 * @param p_timer
 */
void btn_timer_callback (struct k_timer *p_timer)
{
    APP_TASK_EVENTS_T event = EVENT_NONE;

    //
    //! NOTE: Button is active LOW as per schematic
    //! Do not use the device tree pin config, as nrf API can not decode port from the DT
    //
    uint32_t button_pressed = !nrf_gpio_pin_read(GPIO_BUTTON);

    //
    // Process according to button status, we have below button functionalities to handle.
    // Single press
    // Double press
    // Long press - 3secs, for force sleep
    // Long press - 7secs, for factory reset (only of dev testing)
    //
    switch (g_btn_ctx.state) {
        case BTN_STATE_PRESSED:
            if (!button_pressed) {
                btn_context_reset();
                return;
            }
            g_btn_ctx.state = BTN_STATE_DEBOUNCE;
            break;

        case BTN_STATE_DEBOUNCE:
            if (!button_pressed) {
                btn_context_reset();
                return;
            }
            g_btn_ctx.state = BTN_STATE_SINGLE_PRESS;
            break;

        case BTN_STATE_SINGLE_PRESS:
            if (!button_pressed) {
                g_btn_ctx.state = BTN_STATE_DOUBLE_PRESS_WAIT;
                g_btn_ctx.double_press_wait = 0;
            } else if (++g_btn_ctx.long_press_wait >= TRIPLE_PRESS_WAIT) {
                g_btn_ctx.state = BTN_STATE_LONG_PRESS;
            }
            break;

        case BTN_STATE_DOUBLE_PRESS_WAIT:
            if (button_pressed) {
                g_btn_ctx.state = BTN_STATE_DOUBLE_PRESS;
            } else if (++g_btn_ctx.double_press_wait >= DOUBLE_PRESS_WAIT) {
                event = EVENT_BTN_SINGLE_PRESS;
                btn_context_reset();
            }
            break;

        case BTN_STATE_DOUBLE_PRESS:
            if (!button_pressed) {
                g_btn_ctx.state = BTN_STATE_TRIPLE_PRESS_WAIT;
                g_btn_ctx.triple_press_wait = 0;
            } else if (++g_btn_ctx.long_press_wait >= TRIPLE_PRESS_WAIT) {
                g_btn_ctx.state = BTN_STATE_LONG_PRESS;
            }
            break;

        case BTN_STATE_TRIPLE_PRESS_WAIT:
            if (button_pressed) {
                g_btn_ctx.state = BTN_STATE_TRIPLE_PRESS;
            } else if (++g_btn_ctx.triple_press_wait >= TRIPLE_PRESS_WAIT) {
                event = EVENT_BTN_DOUBLE_PRESS;
                btn_context_reset();
            }
            break;

        case BTN_STATE_TRIPLE_PRESS:
            if (!button_pressed) {
                event = EVENT_BTN_TRIPLE_PRESS;
                btn_context_reset();
            } else if (++g_btn_ctx.long_press_wait >= TRIPLE_PRESS_WAIT) {
                g_btn_ctx.state = BTN_STATE_LONG_PRESS;
            }
            break;
        case BTN_STATE_LONG_PRESS:
        {
            //
            // Based on the long press time, post the respective event
            //
            if (g_btn_ctx.long_press_wait >= RESET_LONG_PRESS_MAX_CNT) {
                event = EVENT_BTN_LONG_PRESS_6SECS;
            }
            else if (!button_pressed) {
                btn_context_reset();
                return;
            }

            g_btn_ctx.long_press_wait++;
        }
        break;

        default:
        {
            //
            // Invalid state, clear context
            //
            btn_context_reset();
        }
        break;
    }

    //
    // Post any button event to main thread for event processing
    //
    if (event != EVENT_NONE) {
        btn_context_reset();
        button_event_handler(event);
    }
    else {
        //
        // Need further attention
        //
        k_timer_start(&btn_debounce_timer, K_MSEC(BTN_DEBOUNCE_TIMEOUT_MS), K_NO_WAIT);
    }
}

/**
 * @brief btn_gpio_irq_callback
 * Button IRQ handler
 *
 * @param dev  - pointer to the device structure
 * @param cb   - driver instance callback
 * @param pins - GPIO pins
 */
void btn_gpio_irq_callback (const struct device *dev, struct gpio_callback *cb, uint32_t pins)
{
    if (g_btn_ctx.state == BTN_STATE_RELEASED) {
        g_btn_ctx.state = BTN_STATE_PRESSED;
        k_timer_start(&btn_debounce_timer, K_MSEC(BTN_DEBOUNCE_TIMEOUT_MS), K_NO_WAIT);
    }
}

/**
 * @brief button_init
 * Initialize button and configure the button interrupt to sense button press
 *
 * @return uint8_t
 */
void button_init (void)
{
    //
    // Configure GPIO and interrupt in device tree
    //
    if (!device_is_ready(button.port)) {
        return;
    }

    if (gpio_pin_configure_dt(&button, GPIO_INPUT)) {
        return;
    }

    if (gpio_pin_interrupt_configure_dt(&button, GPIO_INT_EDGE_TO_ACTIVE)) {
        return;
    }

    gpio_init_callback(&button_cb_data, btn_gpio_irq_callback, BIT(button.pin));
    gpio_add_callback(button.port, &button_cb_data);
    k_timer_init(&btn_debounce_timer, btn_timer_callback, NULL);
    g_btn_ctx.state = BTN_STATE_RELEASED;
}
