/* main.c - Application main entry point */

/*
 * Copyright (c) 2021 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: Apache-2.0
 */

/* Includes --------------------------------------------------------------------------------------*/
#include "app_common.h"
#include "buttons.h"
#include "ble_services_peripheral.h"
#include "batt_adc.h"
#include "app_nvs.h"
#include "app_dm.h"
#include <zephyr/random/random.h>
#include <zephyr/bluetooth/hci.h>
#include <zephyr/mgmt/mcumgr/transport/smp_bt.h>
#include <zephyr/pm/pm.h>
#include <zephyr/pm/device.h>
#include <zephyr/pm/policy.h>
#include <hal/nrf_gpio.h>
#include <zephyr/sys/poweroff.h>
#include <zephyr/sys/util.h>
#include <zephyr/device.h>
#include <zephyr/drivers/gpio.h>

/* Macros ----------------------------------------------------------------------------------------*/
/* Thread macros */
#define APP_STACKSIZE               (2048)
#define APP_THREAD_PRIORITY         K_LOWEST_APPLICATION_THREAD_PRIO

/* Private param ---------------------------------------------------------------------------------*/
uint8_t gu8_devices_connected = 0;

/* Global param ----------------------------------------------------------------------------------*/
/* MSG Q for app_thread */
K_MSGQ_DEFINE(app_msgq, sizeof(app_task_msg_t), 16, 4);

/* Timer definitions */
K_TIMER_DEFINE(conn_interval_timer, NULL, NULL);

K_TIMER_DEFINE(device_timer, NULL, NULL);
K_TIMER_DEFINE(peripheral1_timer, NULL, NULL);
K_TIMER_DEFINE(peripheral2_timer, NULL, NULL);
K_TIMER_DEFINE(peripheral3_timer, NULL, NULL);
K_TIMER_DEFINE(peripheral4_timer, NULL, NULL);
device_timers_t timer_ctx[MAX_BEACONS_SUPPORTED + 1];

K_TIMER_DEFINE(battery_monitor_timer, NULL, NULL);

uint8_t dm_index = 0;
message_context_t msg_ctx;
device_information_t dev_info = {
    .zone_range        = ZONE_1,
    .zone_range_far    = ZONE_1 << 1,
    .zone_range_mid    = ZONE_1 * 1.5,
    .zone_range_near   = ZONE_1,
    .conn_interval     = 0,
    .state             = APP_STATE_IDLE,
    .alert_status      = false,
    .zone_change_count = 0,
    .ack               = 0,
    .rssi_range        = ZONE1_RSSI,
};

uint8_t g_dissociate_id[BT_ADDR_SIZE];
const struct device *const cons = DEVICE_DT_GET(DT_CHOSEN(zephyr_console));

/* Private function prototypes -------------------------------------------------------------------*/
static void timers_init (void);
static void timers_context_init (void);
static void beacon_process_ble_data (uint8_t *p_data, uint8_t len);
static void check_beacon_for_breach (uint8_t index, struct dm_result *p_result);
static void check_rssi_for_breach (uint8_t index, int8_t rssi);
static void start_home_coming (struct dm_result *p_result);
static int8_t calculate_rssi_average (uint8_t index);

/**
 * @brief Initialize all the necessary timers for monitoring beacon and detection device
 *
 */
static void timers_init (void)
{
	k_timer_init(&conn_interval_timer, cb_conn_interval_timer, NULL);
	k_timer_init(&battery_monitor_timer, cb_battery_monitor_timer, NULL);
}

/**
 * @brief Initialize all the necessary timers for monitoring beacon and detection device
 *
 */
static void timers_context_init (void)
{
    uint8_t idx = 0;

    timer_ctx[idx].p_timer    = &device_timer;
    timer_ctx[idx].timeout_cb = cb_current_device_timer;

    idx++;
    timer_ctx[idx].p_timer    = &peripheral1_timer;
    timer_ctx[idx].timeout_cb = cb_peripheral1_timer;

    idx++;
    timer_ctx[idx].p_timer    = &peripheral2_timer;
    timer_ctx[idx].timeout_cb = cb_peripheral2_timer;

    idx++;
    timer_ctx[idx].p_timer    = &peripheral3_timer;
    timer_ctx[idx].timeout_cb = cb_peripheral3_timer;

    idx++;
    timer_ctx[idx].p_timer    = &peripheral4_timer;
    timer_ctx[idx].timeout_cb = cb_peripheral4_timer;

    for (idx = 0; idx < MAX_DEVICE_TIMERS_SUPPORTED; idx++) {
        timer_ctx[idx].timeout_ms = K_MSEC(0);
        timer_ctx[idx].event      = TIMER_EVENT_NONE;
        timer_ctx[idx].is_active  = false;

        k_timer_init(timer_ctx[idx].p_timer, timer_ctx[idx].timeout_cb, NULL);
    }
}

/**
 * @brief Process any BLE data received from the detection device here
 *
 * @param p_data - pointer to the data received over BLE
 * @param len    - length of the received data
 */
static void beacon_process_ble_data (uint8_t *p_data, uint8_t len)
{
    app_task_msg_t msg;
    uint8_t opcode = p_data[BYTE_OFF_OPCODE];

    if (opcode == OPCODE_DUMMY_WRITE) {
        // Set the connection status to connected to indicate the beacon can send any
        // notification to the peer and is ready to pair with the peer
        //
        paired_list[0].conn_status = CONN_STATUS_CONNECTED;

        msg.event = EVENT_UPDATE_STATE;
    	msg.state = APP_STATE_WAIT_CONFIRM;
    	k_msgq_put(&app_msgq, &msg, K_NO_WAIT);
    }
    else if (opcode == OPCODE_CONFIRM_PARING) {
        // User confirmed pairing on the remote device, move to paired state
        //
        paired_list[0].conn_status = CONN_STATUS_CONNECTED;

        msg.event = EVENT_UPDATE_STATE;
    	msg.state = APP_STATE_PAIRED;
    	k_msgq_put(&app_msgq, &msg, K_NO_WAIT);
    }
    else if (dev_info.state > APP_STATE_WAIT_CONFIRM) {
		if (opcode == OPCODE_DM_RANGING) {
            k_timer_start(&conn_interval_timer, K_MSEC(dev_info.conn_interval), K_NO_WAIT);

			/* @NOTE: start the timer before printing because printf & memcpy takes 2ms */
			memcpy(&paired_list[0].rand_seed, &p_data[BYTE_OFF_DATA], len);
		}
		else if (opcode == OPCODE_START_ALERT) {
            if (dev_info.state == APP_STATE_HOMING) {
                return;
            }

            if (!dev_info.silent_mode) {
                buzzer_start_alert_tone();
                led_fast_blink(LED_COLOR_RED);
            }

            dev_info.state = APP_STATE_ALERT_START;
		}
		else if (opcode == OPCODE_STOP_ALERT) {
            /* Do not stop the buzzer if in homing state unless single button click in the state */
            if (dev_info.state == APP_STATE_HOMING) {
                return;
            }

			buzzer_tone_stop();
			led_off();
            dev_info.state = APP_STATE_ZONE_RANGING;
		}
		else if (opcode == OPCODE_SET_ZONE) {
			dev_info.zone_range = p_data[BYTE_OFF_DATA];
            dev_info.zone_range_far = dev_info.zone_range * 2;
            dev_info.zone_range_mid = dev_info.zone_range * 1.5;
            dev_info.zone_range_near = dev_info.zone_range;

			turn_on_zone_indicator_led ();
		}
        else if (opcode == OPCODE_DISSOCIATE) {
           	msg.event = EVENT_DISSOCIATE;
        	k_msgq_put(&app_msgq, &msg, K_NO_WAIT);
        }
        else if (opcode == OPCODE_OTA_FOR_BEACON_DEVICE) {
            app_update_state(APP_STATE_SETUP_FOR_OTA);
        }
        else if (opcode == OPCODE_SET_SILENT_MODE) {
            dev_info.zone_range = ZONE_SILENT;
            dev_info.silent_mode = p_data[BYTE_OFF_DATA];
            turn_on_zone_indicator_led();
        }
        else if (opcode == OPCODE_DISABLE_SILENT_MODE) {
            dev_info.zone_range = app_nvs.param.zone_range;
            dev_info.silent_mode = p_data[BYTE_OFF_DATA];
            turn_on_zone_indicator_led ();
        }
        else if (opcode == OPCODE_FIND_DEVICE && !dev_info.silent_mode) {
            buzzer_start_alert_tone();
            led_fast_blink(LED_COLOR_RED);
            dev_info.state = APP_STATE_FIND_THE_DEVICE;
        }
        else if(opcode == OPCODE_DEVICE_FOUND) {

            buzzer_tone_stop();
			led_off();
            dev_info.state = APP_STATE_ZONE_RANGING;
        }
    }
}

/**
 * @brief check_beacon_for_breach
 * Function check the current beacon for any zone breach
 * If zone breached, alert all the beacons.
 * If found device came back to zone, stop alert
 * NOTE: Function should only be called from the detection device
 *
 * @param index    - beacon device ID in the database
 * @param p_result - pointer to the latest DM result
 */
static void
check_beacon_for_breach (uint8_t index, struct dm_result *p_result)
{
    bool breach_status = false;

    do {
        if (dev_info.silent_mode && p_result->dist_estimates.mcpd.best < ZONE_SILENT &&
            !dev_info.alert_status)
        {
            // Device is in silent mode and is inside the silent zone, no other device is breached
            //
            break;
        }

        if (p_result->dist_estimates.mcpd.best >= dev_info.zone_range) {
            paired_list[index].zone_breach_status = true;

            // If the set zone is SILENT, switch to previous set zone and disable silent mode
            //
            if (dev_info.zone_range == ZONE_SILENT) {
                dev_info.zone_range = app_nvs.param.zone_range;
                app_update_state(APP_STATE_ZONE_SETUP);
            }

			app_update_state(APP_STATE_ALERT_START);
		}
        else if (p_result->dist_estimates.mcpd.best > 0.0f && dev_info.alert_status) {
            // Clear the breach status of the result device. Later, check if any other
            // device is out of zone range
            // If any device, do nothing. Else, stop alert
            //
            paired_list[index].zone_breach_status = false;
            for (uint8_t idx = 0; idx < app_nvs.param.paired_device_count; idx++) {
                if (paired_list[idx].zone_breach_status) {
                    breach_status = true;
                    break;
                }
            }

            if (breach_status) {
                break;
            }

            // If we are here means, there was a device out of zone earlier and is now inside the
            // zone, stop any alert triggered
            //
            dev_info.alert_status = false;
            app_update_state(APP_STATE_ALERT_STOP);
        }
    } while (0);
}

/**
 * @brief check_rssi_for_breach
 * Function check the mobile device for any zone breach
 * If zone breached, alert all the beacons.
 * If found device came back to zone, stop alert
 * NOTE: Function should only be called from the detection device
 *
 * @param rssi
 */
static void
check_rssi_for_breach (uint8_t index, int8_t rssi)
{
    bool breach_status = false;

    do {
        if (paired_list[index].dm_mcpd_best > 0) {
            break;
        }

        if (dev_info.silent_mode && (rssi <= ZONE3_RSSI) && !dev_info.alert_status)
        {
            // Device is in silent mode and is inside the silent zone, no other device is breached
            //
            break;
        }

        printk("[%d]: DM unresponsive, checking RSSI...\n", index);

        if (!dev_info.alert_status && dev_info.rssi_range >= rssi) {
            paired_list[index].zone_breach_status = true;

            // If the set zone is SILENT, switch to previous set zone and disable silent mode
            //
            if (dev_info.zone_range == ZONE_SILENT) {
                dev_info.zone_range = app_nvs.param.zone_range;
                app_update_state(APP_STATE_ZONE_SETUP);
            }

			app_update_state(APP_STATE_ALERT_START);
		}
        else if (dev_info.alert_status && dev_info.rssi_range < rssi) {
            // Clear the breach status of the result device. Later, check if any other
            // device is out of zone range
            // If any device, do nothing. Else, stop alert
            //
            paired_list[index].zone_breach_status = false;
            for (uint8_t idx = 0; idx < app_nvs.param.paired_device_count; idx++) {
                if (paired_list[idx].zone_breach_status) {
                    breach_status = true;
                    break;
                }
            }

            if (breach_status) {
                break;
            }

            // If we are here means, there was a device out of zone earlier and is now inside the
            // zone, stop any alert triggered
            //
            dev_info.alert_status = false;
            app_update_state(APP_STATE_ALERT_STOP);
        }
    } while (0);
}

/**
 * @brief start_home_coming
 * Function start the home coming buzzer based on the current range with the detection device
 * NOTE: Function should only be called from the beacon device
 *
 * @param p_result - pointer to the latest DM result
 */
static void
start_home_coming (struct dm_result *p_result)
{
    if (p_result->dist_estimates.mcpd.best >= dev_info.zone_range_far) {
        buzzer_start_homing_tone_slow();
    }
    else if (p_result->dist_estimates.mcpd.best >= dev_info.zone_range_mid) {
        buzzer_start_homing_tone_mid();
    }
    else if (p_result->dist_estimates.mcpd.best >= dev_info.zone_range_near) {
        buzzer_start_homing_tone_fast();
    }
    else if (dev_info.state != APP_STATE_HOMING) {
        buzzer_tone_stop();
        app_update_state(APP_STATE_ZONE_RANGING);
    }
}

/**
 * @brief calculate_rssi_average
 * Function calculates the average of the given beacon's RSSI samples
 *
 * @param index - index of the beacon in our records
 * @return int8_t
 */
static int8_t
calculate_rssi_average (uint8_t index)
{
    int8_t ret_avg = 0;
    int avg = 0;

    do {
        if (index >= app_nvs.param.paired_device_count) {
            break;
        }

        for (uint8_t sample = 0; sample < MAX_RSSI_SAMPLES_COUNT; sample++) {
            avg += paired_list[index].rssi_avg[sample];
        }

        ret_avg = avg / MAX_RSSI_SAMPLES_COUNT;
    } while (0);

    return ret_avg;
}

/**
 * @brief Set the timer event object
 *
 * @param event
 * @param id
 * @param timeout
 */
void set_timer_event (TIMER_EVENT_T event, timer_id_t id, k_timeout_t timeout)
{
    if (event >= TIMER_EVENT_INVALID || id >= TIMER_ID_INVALID) {
        return;
    }

    k_timer_stop(timer_ctx[id].p_timer);

    timer_ctx[id].event = event;
    timer_ctx[id].timeout_ms = timeout;
    timer_ctx[id].is_active = true;

    k_timer_start(timer_ctx[id].p_timer, timeout, K_NO_WAIT);
}

/**
 * @brief Cancel any timer event
 *
 * @param id
 */
void cancel_timer_event (timer_id_t id)
{
    if (id >= TIMER_ID_INVALID) {
        return;
    }

    timer_ctx[id].event = TIMER_EVENT_NONE;
    timer_ctx[id].timeout_ms = K_MSEC(0);
    timer_ctx[id].is_active = false;

    k_timer_stop(timer_ctx[id].p_timer);
}

/**
 * @brief
 * Timer expiry callback function
 *
 * @param p_timer
 */
void cb_current_device_timer (struct k_timer *p_timer)
{
    static app_task_msg_t msg;

    switch(timer_ctx[CURRENT_DEVICE].event) {
    case TIMER_EVENT_PAIRING:
    case TIMER_EVENT_MODE_CHANGE:
    {
        // Move to IDLE state only if no device paired
        //
        if (app_nvs.param.paired_device_count == 0) {
            msg.event = EVENT_UPDATE_STATE;
            msg.state = APP_STATE_IDLE;
        }
        else {
            msg.event = EVENT_UPDATE_STATE;
            msg.state = APP_STATE_ZONE_RANGING;
        }

        k_msgq_put(&app_msgq, &msg, K_NO_WAIT);
    }
    break;

    case TIMER_EVENT_ZONE_CHANGE:
    {
        if (dev_info.zone_change_count++ == (MAX_ZONE * 2)) {
            dev_info.zone_change_count = 0;
            dev_info.zone_range = app_nvs.param.zone_range;
            dev_info.state = APP_STATE_ZONE_RANGING;
            turn_on_zone_indicator_led ();
            break;
        }
    	else if (dev_info.zone_range == ZONE_1) {
    		dev_info.zone_range = ZONE_2;

    		led_zone_or_mode_change_sequence (LED_COLOR_ZONE_2);
    	}
        else if (dev_info.zone_range == ZONE_2) {
    		dev_info.zone_range = ZONE_3;

    		led_zone_or_mode_change_sequence (LED_COLOR_ZONE_3);
    	}
        else if (dev_info.zone_range == ZONE_3) {
    		dev_info.zone_range = ZONE_SILENT;

    		led_zone_or_mode_change_sequence (LED_COLOR_ZONE_SILENT);
        }
        else if (dev_info.zone_range == ZONE_SILENT) {
    		dev_info.zone_range = ZONE_1;

    		led_zone_or_mode_change_sequence (LED_COLOR_ZONE_1);
    	}

        set_timer_event(TIMER_EVENT_ZONE_CHANGE, CURRENT_DEVICE, K_MSEC(ZONE_CHANGE_TIMEOUT_MSEC));
    }
    break;

    default:
    break;
    }
}

/**
 * @brief
 * Timer expiry callback function
 *
 * @param p_timer
 */
void cb_peripheral1_timer (struct k_timer *p_timer)
{
    if (!timer_ctx[PERIPHERAL_1].is_active) {
        return;
    }

    static app_task_msg_t msg;

    switch(timer_ctx[PERIPHERAL_1].event) {
    case TIMER_EVENT_ASSOCIATION:
    {
        msg.event = EVENT_BT_DISCONNECT;
        msg.device_index = PERIPHERAL_1 - 1;
        k_msgq_put(&app_msgq, &msg, K_NO_WAIT);
    }
    break;

    default:
    break;
    }
}

/**
 * @brief
 * Timer expiry callback function
 *
 * @param p_timer
 */
void cb_peripheral2_timer (struct k_timer *p_timer)
{
    if (!timer_ctx[PERIPHERAL_2].is_active) {
        return;
    }

    static app_task_msg_t msg;

    switch(timer_ctx[PERIPHERAL_2].event) {
    case TIMER_EVENT_ASSOCIATION:
    {
        msg.event = EVENT_BT_DISCONNECT;
        msg.device_index = PERIPHERAL_2 - 1;
        k_msgq_put(&app_msgq, &msg, K_NO_WAIT);
    }
    break;

    default:
    break;
    }
}

/**
 * @brief
 * Timer expiry callback function
 *
 * @param p_timer
 */
void cb_peripheral3_timer (struct k_timer *p_timer)
{
    if (!timer_ctx[PERIPHERAL_3].is_active) {
        return;
    }

    static app_task_msg_t msg;

    switch(timer_ctx[PERIPHERAL_3].event) {
    case TIMER_EVENT_ASSOCIATION:
    {
        msg.event = EVENT_BT_DISCONNECT;
        msg.device_index = PERIPHERAL_3 - 1;
        k_msgq_put(&app_msgq, &msg, K_NO_WAIT);
    }
    break;

    default:
    break;
    }
}

/**
 * @brief
 * Timer expiry callback function
 *
 * @param p_timer
 */
void cb_peripheral4_timer (struct k_timer *p_timer)
{
    if (!timer_ctx[PERIPHERAL_4].is_active) {
        return;
    }

    static app_task_msg_t msg;

    switch(timer_ctx[PERIPHERAL_4].event) {
    case TIMER_EVENT_ASSOCIATION:
    {
        msg.event = EVENT_BT_DISCONNECT;
        msg.device_index = PERIPHERAL_4 - 1;
        k_msgq_put(&app_msgq, &msg, K_NO_WAIT);
    }
    break;

    default:
    break;
    }
}

/**
 * @brief Timer expiry callback function for connection interval timer
 *
 * @param p_timer
 */
void cb_conn_interval_timer (struct k_timer *p_timer)
{
	app_task_msg_t msg;
	msg.event = EVENT_START_DM;
	msg.data = NULL;
	k_msgq_put(&app_msgq, &msg, K_NO_WAIT);
}

/**
 * @brief cb_battery_monitor_timer
 * Timer expiry callback function for monitoring battery level
 *
 * @param p_timer
 */
void cb_battery_monitor_timer (struct k_timer *p_timer)
{
	app_task_msg_t msg;
	msg.event = EVENT_BATTERY_MONITOR;
	msg.data = NULL;
	k_msgq_put(&app_msgq, &msg, K_NO_WAIT);
}

/** @brief Update app state. Updates the state flag and kicks the relevant
 *         action including the led display relevant for the state.
 *
 * @param new_state New application state to change to.
 *
 * @return as with k_work_submit_to_queue().
 */
void app_update_state (APP_STATE_T new_state)
{
    /* Homing feature dominates any other state */
    if (dev_info.state == APP_STATE_HOMING) {
        return;
    }

    static app_task_msg_t msg;
	dev_info.state = new_state;

	if (app_nvs.param.mode == DEVICE_MODE_DETECTION) {
		switch (new_state) {
        case APP_STATE_MODE_TRANSIT:
        case APP_STATE_FACTORY_RESET:
        {
            // Mode transit state: request to change the device mode from beacon to detection or
            // vice versa. Remove paring info if any and clear all the device settings and wait for
            // mode transit or new pair with the device.
            //
			msg_ctx.opcode = OPCODE_DISSOCIATE;
			msg_ctx.len	   = 0;
			central_send_msg_to_all_paired_devices(&msg_ctx);

            // We are about to reset device or switch mode, stop all the operations, reset all the
            // states and prepare for switching mode
            //
            k_timer_stop(&dm_timer);
            stop_scan();
            buzzer_tone_stop();

            memset(&paired_list, 0, sizeof(paired_list));

            // Clear NVS data
            // NOTE: Retain device mode
            //
            DEVICE_MODE_T mode = app_nvs.param.mode;
            memset(&app_nvs, 0, sizeof(app_nvs));
            app_nvs.param.mode = mode;
            app_nvs.param.zone_range = ZONE_1;
            update_nvs();

            // Show device is in transit mode using LED indications
            //
            led_zone_or_mode_change_sequence(LED_COLOR_WHITE);
            k_timer_start(&led_timer, K_MSEC(MODE_CHANGE_TIMEOUT_MSEC), K_NO_WAIT);
        }
        break;

		case APP_STATE_IDLE:
        {
            if ((!app_nvs.param.paired_device_count) ||
                (app_nvs.param.paired_device_count >= MAX_BEACONS_SUPPORTED))
            {
                led_off();
                stop_scan();
            }
        }
		break;

		case APP_STATE_PAIRING:
        {
			led_slow_blink(LED_COLOR_PALE_BLUE);
			start_scan();
            set_timer_event(TIMER_EVENT_PAIRING, CURRENT_DEVICE, K_MSEC(PAIRING_TIMEOUT_MSEC));
        }
		break;

        case APP_STATE_GET_DEVICE_ID:
        {
            // The mobile BLE address is not static, so mobile devices are paired with mobile ID
            // which is unique and set from the mobile app.
            // Request for the ID to check for existence
            // For other notification devices, check from the BLE address which are stored in our
            // database
            //
            if (current_dev_info.is_mobile) {
                uint8_t opcode = OPCODE_GET_MOBILE_ID;

                static struct bt_gatt_write_params params;
                params.handle = current_dev_info.cmd_chr_handle;
                params.data   = &opcode;
                params.length = sizeof(opcode);
                params.func   = write_resp_cb;

                bt_gatt_write(current_dev_info.conn, &params);
            }
            else {
                // Check for existence device
                //
                bt_addr_le_t dev_addr;
                memcpy(&dev_addr, bt_conn_get_dst(current_dev_info.conn), sizeof(bt_addr_le_t));
                memcpy(current_dev_info.id, dev_addr.a.val, sizeof(current_dev_info.id));
                current_dev_info.addr = dev_addr;
                msg.event = EVENT_UPDATE_STATE;
			    msg.state = APP_STATE_CHECK_EXISTENCE;
			    k_msgq_put(&app_msgq, &msg, K_NO_WAIT);
            }
        }
        break;

        case APP_STATE_CHECK_EXISTENCE:
        {
            // Check if the connected device is already paired with us.
            // For new device, assign a slot for storing all the device info
            // For paired device, update device info as connected.
            //
            int8_t device_idx = get_paired_device_slot_from_id_param(current_dev_info.id);
            if (device_idx >= 0) {
                g_current_index = device_idx;
                paired_list[g_current_index] = current_dev_info;
        		paired_list[g_current_index].conn_status = CONN_STATUS_CONNECTED;
				msg.event = EVENT_PAIRED_DEVICE_CONNECTED;
                msg.device_index = g_current_index;
            	k_msgq_put(&app_msgq, &msg, K_NO_WAIT);
			}
            else if (app_nvs.param.paired_device_count < MAX_BEACONS_SUPPORTED) {
                g_current_index = app_nvs.param.paired_device_count;
                paired_list[g_current_index] = current_dev_info;
                msg.event = EVENT_UPDATE_STATE;
                msg.state = APP_STATE_WAIT_CONFIRM;
                msg.device_index = g_current_index;
                k_msgq_put(&app_msgq, &msg, K_NO_WAIT);
			}
            else {
                // No room for new connection, disconnect
                //
                bt_conn_disconnect(current_dev_info.conn, BT_HCI_ERR_REMOTE_USER_TERM_CONN);
            }
        }
        break;

		case APP_STATE_WAIT_CONFIRM:
		{
		    led_fast_blink(LED_COLOR_GREEN);
            cancel_timer_event(CURRENT_DEVICE);
            set_timer_event(TIMER_EVENT_ASSOCIATION, (g_current_index + 1),
                            K_MSEC(ASSOCIATION_TIMEOUT_MSEC));

            // Write dummy data to indicate beacon, the detection device is ready to read indication
            //
            uint8_t opcode = OPCODE_DUMMY_WRITE;
            app_gatt_write(g_current_index, &opcode, sizeof(opcode));
		}
		break;

		case APP_STATE_PAIRED:
		{
			cancel_timer_event(g_current_index + 1);

            /* Since new device is connected, config the zone range to the new device */
            uint8_t data[16];
			data[BYTE_OFF_OPCODE] = OPCODE_SET_ZONE;
			data[BYTE_OFF_LEN]	  = sizeof(dev_info.zone_range);
			data[BYTE_OFF_DATA]	  = dev_info.zone_range;
			app_gatt_write(g_current_index, data, data[BYTE_OFF_LEN] + BLE_PKT_HEADER_SZ);

			paired_list[g_current_index].pair_status = true;
            paired_list[g_current_index].conn_status = CONN_STATUS_CONNECTED;
			update_nvs_connection_param(paired_list[g_current_index].id);

			led_set_color(LED_COLOR_GREEN);
			k_timer_start(&led_timer, K_SECONDS(LED_TIMEOUT_SEC), K_NO_WAIT);

            if (dev_info.alert_status) {
				msg.event = EVENT_UPDATE_STATE;
				msg.state = APP_STATE_ALERT_STOP;
                k_msgq_put(&app_msgq, &msg, K_NO_WAIT);
            }

            dev_info.state = APP_STATE_ZONE_RANGING;

            // Start scanning for other beacon devices
            //
            k_msleep(200);
            start_scan();

            // We only need to start the DM timer at once when the first device is paired or paired
            // device is reconnected after DETECTION device is disconnected from the network.
            // So it is better to monitor device connected in the session and start DM timer
            // accordingly
            //
            gu8_devices_connected++;
            if (gu8_devices_connected == 1) {
                // Paired with the first device, start timer for DM
                //
                k_timer_start(&dm_timer, K_MSEC(DM_TIME_INTERVAL_MSEC), K_NO_WAIT);
            }
		}
		break;

		case APP_STATE_ZONE_SETUP:
		{
			// Pack the data to send over BLE
			//
            msg_ctx.opcode  = OPCODE_SET_ZONE;
			msg_ctx.data[0] = dev_info.zone_range;
			msg_ctx.len	    = sizeof(dev_info.zone_range);

            if (dev_info.zone_range == ZONE_1) {
                dev_info.rssi_range = ZONE1_RSSI;
            }
            else if (dev_info.zone_range == ZONE_2) {
                dev_info.rssi_range = ZONE2_RSSI;
            }
            else if (dev_info.zone_range == ZONE_3) {
                dev_info.rssi_range = ZONE3_RSSI;
            }
            else if (dev_info.zone_range == ZONE_SILENT) {
                dev_info.rssi_range = ZONE_SILENT_RSSI;
            }

            central_send_msg_to_all_paired_devices(&msg_ctx);
			led_off();

            if (dev_info.zone_range == ZONE_SILENT) {
                // If the set zone is SILENT ZONE, also set silent mode
                //
                dev_info.silent_mode = true;
            }
            else {
                app_nvs.param.zone_range = dev_info.zone_range;
                dev_info.silent_mode = false;
            }

            update_nvs();

			dev_info.state = APP_STATE_ZONE_RANGING;
		}
		break;

		case APP_STATE_ALERT_START:
		{
            if (!dev_info.silent_mode) {
			    buzzer_start_alert_tone();
			    led_fast_blink(LED_COLOR_RED);
            }

			// Pack the data to send over BLE
			//
            msg_ctx.opcode = OPCODE_START_ALERT;
            msg_ctx.len	   = 0;
            central_send_msg_to_all_paired_devices(&msg_ctx);

            dev_info.alert_status = true;
		}
		break;

		case APP_STATE_ALERT_STOP:
		{
			// Pack the data to send over BLE
			//
			msg_ctx.opcode = OPCODE_STOP_ALERT;
			msg_ctx.len	   = 0;
			central_send_msg_to_all_paired_devices(&msg_ctx);
            buzzer_tone_stop();
            led_off();
            dev_info.state = APP_STATE_ZONE_RANGING;
		}
		break;

        case APP_STATE_SETUP_FOR_OTA:
        {
            // Setup detection device for OTA, disconnect from all the beacons
            // Stop scanning and start advertisement to receive OTA from the app
            //
            for (uint8_t idx = 0; idx < app_nvs.param.paired_device_count; idx++) {
				if (paired_list[idx].conn_status != CONN_STATUS_DISCONNECTED &&
                    paired_list[idx].conn)
                {
					bt_conn_disconnect(paired_list[idx].conn, BT_HCI_ERR_REMOTE_USER_TERM_CONN);
                    if (paired_list[idx].conn) {
                        bt_conn_unref(paired_list[idx].conn);
                        paired_list[idx].conn = NULL;
                    }
				}
            }

            stop_scan();
			advertisement_start();
            led_slow_blink(LED_COLOR_PALE_BLUE);
            dev_info.state = APP_STATE_OTA_UPDATE;
        }
        break;

        case APP_STATE_OTA_UPDATE:
        {
            /* Do nothing for now */
        }
        break;

        case APP_STATE_SETUP_BEACON_FOR_OTA:
        {
            if (app_nvs.param.mode == DEVICE_MODE_BEACON) {
                break;
            }

            uint8_t opcode = OPCODE_OTA_FOR_BEACON_DEVICE;

            /* Setup all beacons for OTA update one after the other */
            for (uint8_t idx = 0; idx < app_nvs.param.paired_device_count; idx++) {
                if (paired_list[idx].fw_version_major == FW_VERSION_MAJOR &&
                    paired_list[idx].fw_version_minor == FW_VERSION_MINOR)
                {
                    continue;
                }
                app_gatt_write(idx, &opcode, sizeof(opcode));
                break;
            }
        }
        break;

        case APP_STATE_HOMING:
        {
            /* No operation in detection device */
        }
        break;

        case APP_STATE_SILENT_MODE:
        {
            dev_info.silent_mode = 1;
			msg_ctx.opcode = OPCODE_SET_SILENT_MODE;
			msg_ctx.len	   = sizeof(dev_info.silent_mode);
            msg_ctx.data[BYTE_OFF_DATA] = dev_info.silent_mode;
			central_send_msg_to_all_paired_devices(&msg_ctx);
            dev_info.state = APP_STATE_SILENT_MODE;
        }
        break;

        case APP_STATE_DISABLE_SILENT_MODE:
        {
            dev_info.silent_mode = 0;
            msg_ctx.opcode = OPCODE_DISABLE_SILENT_MODE;
			msg_ctx.len	   = sizeof(dev_info.silent_mode);
            msg_ctx.data[BYTE_OFF_DATA] = dev_info.silent_mode;
			central_send_msg_to_all_paired_devices(&msg_ctx);
            turn_on_zone_indicator_led();
            dev_info.state = APP_STATE_ZONE_RANGING;
        }
        break;

        case APP_STATE_FIND_THE_DEVICE:
        {
		    if (!dev_info.silent_mode) {
                buzzer_start_alert_tone();
                led_fast_blink(LED_COLOR_RED);
            }

			// Pack the data to send over BLE
			//
            msg_ctx.opcode = OPCODE_FIND_DEVICE;
            msg_ctx.len	   = 0;
            central_send_msg_to_all_paired_devices(&msg_ctx);

            dev_info.state = APP_STATE_FIND_THE_DEVICE;
        }
        break;

        case APP_STATE_DEVICE_FOUND:
		{
            buzzer_tone_stop();
            led_off();

			// Pack the data to send over BLE
			//
			msg_ctx.opcode = OPCODE_DEVICE_FOUND;
			msg_ctx.len	   = 0;
			central_send_msg_to_all_paired_devices(&msg_ctx);

            dev_info.state = APP_STATE_ZONE_RANGING;
		}
		break;

		default:
		break;
		}
	}
    else {
        /* Beacon events */
		switch (new_state) {
		case APP_STATE_MODE_TRANSIT:
        case APP_STATE_FACTORY_RESET:
        {
			// Pack the data to send over BLE
			//
			uint8_t opcode = OPCODE_DISSOCIATE;
			bt_indicate(&opcode, sizeof(opcode));

            memset(&paired_list, 0, sizeof(paired_list));

            // Turn OFF buzzer just in case
            //
            buzzer_tone_stop();

            // Clear NVS data
            // !!! NOTE: Retain device mode !!!
            //
            DEVICE_MODE_T mode = app_nvs.param.mode;
            memset(&app_nvs, 0, sizeof(app_nvs));
            app_nvs.param.mode = mode;
            app_nvs.param.zone_range = ZONE_1;
            update_nvs();

            led_zone_or_mode_change_sequence(LED_COLOR_WHITE);
            k_timer_start(&led_timer, K_MSEC(MODE_CHANGE_TIMEOUT_MSEC), K_NO_WAIT);
        }
		break;

		case APP_STATE_IDLE:
        {
            advertisement_stop();
			led_off();
            buzzer_tone_stop();
        }
		break;

		case APP_STATE_PAIRING:
        {
			led_slow_blink(LED_COLOR_GREEN);
			advertisement_start();
            set_timer_event(TIMER_EVENT_PAIRING, CURRENT_DEVICE, K_MSEC(PAIRING_TIMEOUT_MSEC));
        }
		break;

		case APP_STATE_WAIT_CONFIRM:
		{
			led_fast_blink(LED_COLOR_GREEN);
            cancel_timer_event(CURRENT_DEVICE);
		}
		break;

		case APP_STATE_PAIRED:
        {
			led_set_color(LED_COLOR_GREEN);
			k_timer_start(&led_timer, K_SECONDS(LED_TIMEOUT_SEC), K_NO_WAIT);

            /* If we are paired with the peer earlier and reconnected, stop the alert */
		    paired_list[0].conn_status = CONN_STATUS_CONNECTED;
            if (paired_list[0].pair_status) {
                buzzer_tone_stop();
            }
            else {
                bt_addr_le_t dev_addr;
                memcpy(&dev_addr, bt_conn_get_dst(paired_list[0].conn), sizeof(bt_addr_le_t));
                memcpy(paired_list[0].id, dev_addr.a.val, sizeof(paired_list[0].id));
			    paired_list[0].pair_status = true;
			    update_nvs_connection_param(paired_list[0].id);
            }

            /* Update FW info to the detection device */
            uint8_t fw_info[] = {OPCODE_FW_VER_INFO, 2, FW_VERSION_MAJOR, FW_VERSION_MINOR};
            bt_indicate(fw_info, sizeof(fw_info));

            dev_info.state = APP_STATE_ZONE_RANGING;
        }
		break;

		case APP_STATE_ZONE_SETUP:
		{
			uint8_t data[3];
            data[BYTE_OFF_OPCODE] = OPCODE_SET_ZONE;
			data[BYTE_OFF_LEN]    = sizeof(dev_info.zone_range);
			data[BYTE_OFF_DATA]   = dev_info.zone_range;
			bt_indicate(data, sizeof(data));
			led_off();

            dev_info.zone_range_far = dev_info.zone_range * 2;
            dev_info.zone_range_mid = dev_info.zone_range * 1.5;
            dev_info.zone_range_near = dev_info.zone_range + 1;

            if (dev_info.zone_range == ZONE_SILENT) {
                // If the set zone is SILENT ZONE, also set silent mode
                //
                dev_info.silent_mode = true;
            }
            else {
                app_nvs.param.zone_range = dev_info.zone_range;
                dev_info.silent_mode = false;
            }

            update_nvs();

			dev_info.state = APP_STATE_ZONE_RANGING;
		}
		break;

		case APP_STATE_ALERT_START:
		{

		    if (!dev_info.silent_mode) {
                buzzer_start_alert_tone();
                led_fast_blink(LED_COLOR_RED);
            }

            // If detection device disconnected and we are here, start the advertisement and let the
            // beacon alert
            //
            if (paired_list[0].conn_status != CONN_STATUS_CONNECTED) {
    			advertisement_start();
            }
		}
		break;

		case APP_STATE_ALERT_STOP:
		{
			led_off();
            buzzer_tone_stop();

			uint8_t data[3];
			data[BYTE_OFF_OPCODE] = OPCODE_STOP_ALERT;
			data[BYTE_OFF_LEN]    = 0;
			data[BYTE_OFF_DATA]   = 0;
			bt_indicate(data, sizeof(data));

            // If detection device is disconnected and we are here, move to pairing
            //
            if (paired_list[0].conn_status != CONN_STATUS_CONNECTED) {
                dev_info.state = APP_STATE_PAIRING;
                break;
            }

            // Comms are good, monitor zone
            //
			dev_info.state = APP_STATE_ZONE_RANGING;
		}
		break;

        case APP_STATE_SETUP_FOR_OTA:
        {
            // Disconnect from detection device and connect to mobile for OTA
            //
			bt_conn_disconnect(paired_list[0].conn, BT_HCI_ERR_REMOTE_USER_TERM_CONN);
			paired_list[0].conn = NULL;

			advertisement_start();
            led_slow_blink(LED_COLOR_GREEN);
            dev_info.state = APP_STATE_OTA_UPDATE;
        }
        break;

        case APP_STATE_OTA_UPDATE:
        {
            /* Do nothing for now */
        }
        break;

        case APP_STATE_HOMING:
        {
            buzzer_tone_stop();
            led_off();
            buzzer_start_homing_tone_mid();
        }
        break;

        case APP_STATE_SILENT_MODE:
        {
            uint8_t data[3];
			data[BYTE_OFF_OPCODE] = OPCODE_SET_SILENT_MODE;
			data[BYTE_OFF_LEN]    = sizeof(dev_info.silent_mode);
			data[BYTE_OFF_DATA]   = 1;
			bt_indicate(data, sizeof(data));
            dev_info.state = APP_STATE_SILENT_MODE;

        }
        break;

        case APP_STATE_DISABLE_SILENT_MODE:
        {
            uint8_t data[3];
			data[BYTE_OFF_OPCODE] = OPCODE_DISABLE_SILENT_MODE;
			data[BYTE_OFF_LEN]    = sizeof(dev_info.silent_mode);
			data[BYTE_OFF_DATA]   = 0;
			bt_indicate(data, sizeof(data));
            dev_info.state = APP_STATE_ZONE_RANGING;
            turn_on_zone_indicator_led();
        }
        break;

        case APP_STATE_FIND_THE_DEVICE:
        {
           if (!dev_info.silent_mode) {
                buzzer_start_alert_tone();
                led_fast_blink(LED_COLOR_RED);

                // Send the indication regarding finding the device over BLE
                //
                uint8_t data[3];
    			data[BYTE_OFF_OPCODE] = OPCODE_FIND_DEVICE;
    			data[BYTE_OFF_LEN]    = 0;
    			data[BYTE_OFF_DATA]   = 0;
    			bt_indicate(data, sizeof(data));
            }
        }
        break;

        case APP_STATE_DEVICE_FOUND:
        {
            buzzer_tone_stop();
            led_off();

            // Send the indication regarding status of device found the device over BLE
            //
            uint8_t data[3];
			data[BYTE_OFF_OPCODE] = OPCODE_DEVICE_FOUND;
			data[BYTE_OFF_LEN]    = 0;
			data[BYTE_OFF_DATA]   = 0;
			bt_indicate(data, sizeof(data));
        }
        break;

		default:
		break;
		}
	}
}

/**
 * @brief Thread for handling application events
 *
 */
void app_thread (void)
{
    app_task_msg_t msg;

	while (1) {
        /* Wait for any event */
		if (k_msgq_get(&app_msgq, &msg, K_FOREVER)== 0) {
            switch (msg.event) {
            case EVENT_BTN_SINGLE_PRESS:
            {
                if (dev_info.state == APP_STATE_SILENT_MODE) {
                    led_zone_or_mode_change_sequence(LED_COLOR_ZONE_SILENT);
                }
                else if (app_nvs.param.mode == DEVICE_MODE_DETECTION) {
                    if (dev_info.state == APP_STATE_IDLE) {
                        app_update_state(APP_STATE_PAIRING);
                    }
                    else if (dev_info.state == APP_STATE_PAIRING) {
                        led_fast_blink(LED_COLOR_GREEN);
                        k_timer_start(&led_timer, K_MSEC(500), K_NO_WAIT);
                    }
                    else if (dev_info.state == APP_STATE_WAIT_CONFIRM) {
						/* Let the remote device know that the pairing has been confirmed
						   by user on this device */
						uint8_t opcode = OPCODE_CONFIRM_PARING;
						app_gatt_write(g_current_index, &opcode, sizeof(opcode));
                        app_update_state(APP_STATE_PAIRED);
                    }
                    else if (dev_info.state == APP_STATE_ZONE_RANGING) {
                        app_update_state(APP_STATE_FIND_THE_DEVICE);
                    }
                    else if (dev_info.state == APP_STATE_ZONE_SETUP) {
                        /* Stop the on going timer as user has confirmed the zone and also clear any
                        timer event */
                        dev_info.zone_change_count = 0;
                        k_timer_stop(timer_ctx[CURRENT_DEVICE].p_timer);
						led_off();
                        app_update_state(APP_STATE_ZONE_SETUP);
                        turn_on_zone_indicator_led();
                    }
                    else if (dev_info.state == APP_STATE_ALERT_START) {
                        dev_info.ack = true;
                        app_update_state(APP_STATE_ALERT_STOP);
                    }
                    else if (dev_info.state == APP_STATE_FIND_THE_DEVICE) {
                        app_update_state(APP_STATE_DEVICE_FOUND);
                        turn_on_zone_indicator_led ();
                    }
                }
                else if (app_nvs.param.mode == DEVICE_MODE_BEACON) {
                    if (dev_info.state == APP_STATE_IDLE) {
                        app_update_state(APP_STATE_PAIRING);
                    }
                    else if (dev_info.state == APP_STATE_PAIRING) {
                        if (paired_list[0].conn_status != CONN_STATUS_DISCONNECTED) {
                            led_fast_blink(LED_COLOR_GREEN);
                            k_timer_start(&led_timer, K_MSEC(500), K_NO_WAIT);
                            set_timer_event(TIMER_EVENT_PAIRING, CURRENT_DEVICE,
                                            K_MSEC(PAIRING_TIMEOUT_MSEC));
                        }
                    }
                    else if (dev_info.state == APP_STATE_WAIT_CONFIRM) {
                        /* User has confirm pairing, send the same notification to the peer */
                        uint8_t opcode = OPCODE_CONFIRM_PARING;
                        bt_indicate(&opcode, sizeof(opcode));

                        app_update_state(APP_STATE_PAIRED);
                    }
                    else if (dev_info.state == APP_STATE_ZONE_RANGING) {
                        app_update_state(APP_STATE_FIND_THE_DEVICE);
                    }
                    else if (dev_info.state == APP_STATE_ZONE_SETUP) {
                        dev_info.zone_change_count = 0;
                        cancel_timer_event(CURRENT_DEVICE);
						led_off();
                        app_update_state(APP_STATE_ZONE_SETUP);
                        turn_on_zone_indicator_led();
                    }
                    else if(dev_info.state == APP_STATE_ALERT_START) {
                        app_update_state(APP_STATE_ALERT_STOP);
                    }
                    else if (dev_info.state == APP_STATE_HOMING) {
                        buzzer_tone_stop();
                        dev_info.state = APP_STATE_ZONE_RANGING;
                    }
                    else if(dev_info.state == APP_STATE_FIND_THE_DEVICE){
                        app_update_state(APP_STATE_DEVICE_FOUND);
                        turn_on_zone_indicator_led ();
                    }
                }
                else {
                    /* Should not come here */
                }
            }
            break;

            case EVENT_BTN_DOUBLE_PRESS:
            {
                if (dev_info.state == APP_STATE_MODE_TRANSIT) {
                    cancel_timer_event(CURRENT_DEVICE);

                    led_fast_blink(LED_COLOR_WHITE);
                    k_timer_start(&led_timer, K_MSEC(500), K_NO_WAIT);

                    /* Switch the device mode and reboot to take effect */
                    app_nvs.param.mode = ((app_nvs.param.mode == DEVICE_MODE_DETECTION) ?
                                          DEVICE_MODE_BEACON : DEVICE_MODE_DETECTION);
					update_nvs();

                    k_sleep(K_SECONDS(1));
                    sys_reboot(0);
                }
                else if (dev_info.state >= APP_STATE_PAIRED) {
                    if (app_nvs.param.mode == DEVICE_MODE_BEACON &&
                        dev_info.state == APP_STATE_ALERT_START)
                    {
                        dev_info.state = APP_STATE_HOMING;
                        app_update_state(APP_STATE_HOMING);
                    }
                    else if (dev_info.state == APP_STATE_SILENT_MODE) {
                        // If the device is in silent mode update the device to previous state and
                        // previous ZONE RANGE
                        //
                        app_update_state(APP_STATE_DISABLE_SILENT_MODE);
                    }
                    else {
                        set_timer_event(TIMER_EVENT_ZONE_CHANGE, CURRENT_DEVICE, K_MSEC(1));
    					dev_info.state = APP_STATE_ZONE_SETUP;
                    }
                }
            }
            break;

            case EVENT_BTN_TRIPLE_PRESS:
            {
                if (dev_info.state == APP_STATE_PAIRED) {
                    app_update_state(APP_STATE_SILENT_MODE);
                    led_zone_or_mode_change_sequence(LED_COLOR_ZONE_SILENT);
                }
            }
            break;

            case EVENT_BTN_LONG_PRESS_6SECS:
            {
                set_timer_event(TIMER_EVENT_MODE_CHANGE, CURRENT_DEVICE,
                                K_MSEC(MODE_CHANGE_TIMEOUT_MSEC));
                app_update_state(APP_STATE_MODE_TRANSIT);
            }
            break;

			case EVENT_DM_RESULT_READY:
            {
				struct dm_result *p_result = (struct dm_result *)msg.data;
				int8_t index = get_paired_device_slot_from_id_param(p_result->bt_addr.a.val);
                if (index < 0) {
                    break;
                }

                printk("[%d]: Measurement result:\n", index);
                print_result(p_result);

                uint16_t handle = 0;
                if (bt_hci_get_conn_handle(paired_list[index].conn, &handle) == 0) {
                    int8_t tx_power = 0;
                    read_conn_tx_power(handle, &tx_power);
                    printk("\tTX: %d\n\n", tx_power);

                    if (tx_power < MAX_TX_POWER) {
                        set_tx_power(handle, MAX_TX_POWER);
                    }
                }

				if (app_nvs.param.mode == DEVICE_MODE_DETECTION) {
                    paired_list[index].dm_mcpd_best = p_result->dist_estimates.mcpd.best;
                    check_beacon_for_breach(index, p_result);
                }
                else if (app_nvs.param.mode == DEVICE_MODE_BEACON &&
                         dev_info.state == APP_STATE_HOMING)
                {
                    start_home_coming(p_result);
                }
            }
			break;

			case EVENT_START_DM:
            {
                int8_t idx = *((int8_t *)msg.data);

				if (app_nvs.param.mode == DEVICE_MODE_DETECTION) {
					if (paired_list[idx].conn_status == CONN_STATUS_CONNECTED &&
						paired_list[idx].pair_status == 1 && !paired_list[idx].is_mobile)
                    {
						detection_dm_add_request(paired_list[idx].rand_seed, paired_list[idx].addr);
						printk("[%d]: Dev seed: %x\n", idx, paired_list[idx].rand_seed);

                        // Since we have added DM request, clear previous DM estimated value
                        // This way detection device can know whether the DM result is being updated
                        // regularly or not.
                        // If not, detection device can check RSSI for DM estimation
                        //
                        paired_list[idx].dm_mcpd_best = 0;
					}
				}
				else if (app_nvs.param.mode == DEVICE_MODE_BEACON) {
					beacon_dm_add_request(paired_list[0].rand_seed, paired_list[0].addr);
					printk("[0]: Dev seed: %x\n", paired_list[0].rand_seed);
				}
            }
			break;

            case EVENT_BT_DISCONNECT:
            {
                if (app_nvs.param.mode == DEVICE_MODE_DETECTION) {
			        cancel_timer_event(msg.device_index + 1);

                    int8_t device_index = msg.device_index;
                    if (device_index == -1) {
                        start_scan();   // scan for fresh or other devices
                        break;
                    }

    				if (paired_list[device_index].conn_status != CONN_STATUS_DISCONNECTED &&
                        paired_list[device_index].conn)
                    {
    					bt_conn_disconnect(paired_list[device_index].conn,
                                           BT_HCI_ERR_REMOTE_USER_TERM_CONN);
                        if (paired_list[device_index].conn) {
                            bt_conn_unref(paired_list[device_index].conn);
                            paired_list[device_index].conn = NULL;
                        }
    				}

                    /* If any device is paired, continue to monitor distance. Else, be in IDLE */
                    if (app_nvs.param.paired_device_count) {
                        start_scan();
                    }
                    else {
                        app_update_state(APP_STATE_IDLE);
                    }
                }
                else {
    				if (paired_list[0].conn_status != CONN_STATUS_DISCONNECTED) {
    					bt_conn_disconnect(paired_list[0].conn, BT_HCI_ERR_AUTH_FAIL);
    					paired_list[0].conn = NULL;
                        paired_list[0].conn_status = CONN_STATUS_DISCONNECTED;
    				}

                    if (!paired_list[0].pair_status) {
    				    app_update_state(APP_STATE_IDLE);
                    }
                }
            }
            break;

            case EVENT_UPDATE_NVS:
            {
                update_nvs();
            }
            break;

            case EVENT_UPDATE_STATE:
            {
                app_update_state(msg.state);
            }
            break;

			case EVENT_PERIODIC_DM_REQ:
			{
                // NOTE: DM req should be added from the detection device
                //
                if (app_nvs.param.mode == DEVICE_MODE_BEACON) {
                    break;
                }

                static uint8_t dm_index = 0;
                bool b_is_not_mobile = false;
                uint16_t handle = 0;

                // Check current peer connection status,
                // If not connected, skip to next device in iteration
                // NOTE: Since DM is NRF application, read the RSSI for mobiles
                //
                for ( ; dm_index < app_nvs.param.paired_device_count; dm_index++) {
                    if (paired_list[dm_index].conn_status != CONN_STATUS_CONNECTED) {
                        continue;
                    }

#if ENABLE_DM
                    if (!paired_list[dm_index].is_mobile) {
                        b_is_not_mobile = true;
                    }
#endif

                    if (bt_hci_get_conn_handle(paired_list[dm_index].conn, &handle) == 0) {
                        uint8_t off = paired_list[dm_index].rssi_off;
                        read_conn_rssi(handle, &paired_list[dm_index].rssi_avg[off]);
                        paired_list[dm_index].rssi_off = ((off < (MAX_RSSI_SAMPLES_COUNT - 1)) ?
                                                          (off + 1) : 0);
                        int8_t avg_rssi = calculate_rssi_average(dm_index);
                        printk("[%d]: RSSI: %d\n", dm_index, avg_rssi);
                        check_rssi_for_breach(dm_index, avg_rssi);

                        break; // for(;;) loop break
                    }
                }

                if (b_is_not_mobile) {
    				uint8_t data[6];

    				paired_list[dm_index].rand_seed = sys_rand32_get();
    				data[BYTE_OFF_OPCODE] = OPCODE_DM_RANGING;
    				data[BYTE_OFF_LEN]    = sizeof(paired_list[dm_index].rand_seed);
                    memcpy(&data[BYTE_OFF_DATA], &paired_list[dm_index].rand_seed,
                           sizeof(paired_list[dm_index].rand_seed));
    				app_gatt_write(dm_index, data, sizeof(data));
                    gb_dm_write_resp_wait = true;
                }

                dm_index = ((dm_index < (app_nvs.param.paired_device_count - 1)) ?
                            (dm_index + 1) : 0);

                k_timer_start(&dm_timer, K_MSEC(DM_TIME_INTERVAL_MSEC), K_NO_WAIT);
			}
            break;

            case EVENT_PROCESS_BLE_DATA:
            {
                uint8_t *p_data = (uint8_t *)msg.data;
                beacon_process_ble_data(msg.data, p_data[BYTE_OFF_LEN]);
            }
            break;

            case EVENT_PAIRED_DEVICE_CONNECTED:
            {
				/* Let the remote device know that the pairing has been confirmed
				   by user on this device */
				uint8_t data[6];

                if (app_nvs.param.mode == DEVICE_MODE_DETECTION &&
                    paired_list[g_current_index].conn_status == CONN_STATUS_CONNECTED)
                {
    				data[BYTE_OFF_OPCODE] = OPCODE_CONFIRM_PARING;
    				app_gatt_write(msg.device_index, data, 1);
                    app_update_state(APP_STATE_PAIRED);
                }
            }
            break;

            case EVENT_DISSOCIATE:
            {
                bool status_ok = false;

                if (app_nvs.param.mode == DEVICE_MODE_DETECTION) {
                    /* Get the device slot index from the pair info before removing/deleting from
                    the paired list */
                    int8_t dev_idx = (*(int8_t *)msg.data);
                    if (dev_idx == -1) {
                        break;
                    }

                    status_ok = remove_paired_dev(dev_idx);
                    if (!status_ok) {
                        break;
                    }

                    /* Rearrange the device connection info if we have multiple devices */
                    if (app_nvs.param.paired_device_count) {
                        for ( ; dev_idx < (app_nvs.param.paired_device_count - 1); dev_idx++) {
                            memcpy(app_nvs.param.pair_info[dev_idx].id,
                                   app_nvs.param.pair_info[dev_idx + 1].id,
                                   sizeof(app_nvs.param.pair_info[dev_idx].id));
                        }

                        memset(&app_nvs.param.pair_info[dev_idx].id, 0,
                               sizeof(app_nvs.param.pair_info[dev_idx].id));
                        --app_nvs.param.paired_device_count;
                    }

                    if (!app_nvs.param.paired_device_count) {
                        app_update_state(APP_STATE_IDLE);
                    }
                }
                else {
                    status_ok = remove_paired_dev(0);
                    if (!status_ok) {
                        break;
                    }

                    memset((void *)&app_nvs, 0, sizeof(app_nvs));
                    app_nvs.param.mode = DEVICE_MODE_BEACON;

                    /* Since we are dissociated with the detection device, move idle state */
                    app_update_state(APP_STATE_IDLE);
                }

                update_nvs();
            }
            break;

            case EVENT_BATTERY_MONITOR:
            {
                if (batt_read() <= BATTERY_LOW_THRESHOLD) {
                    led_slow_blink(LED_COLOR_MAGENTA);
                    k_timer_start(&led_timer, K_MSEC(5000), K_NO_WAIT);
                }

                k_timer_start(&battery_monitor_timer, K_SECONDS(BATTERY_MONITOR_INTERVAL_SECS),
                              K_NO_WAIT);
            }
            break;

			default:
			break;
            }
        }
	}
}

// Main application thread for handling events
//
K_THREAD_DEFINE(app_thread_id, APP_STACKSIZE, app_thread,
		        NULL, NULL, NULL, APP_THREAD_PRIORITY, 0, 0);

/**
 * @brief main
 * Entry point to the application
 *
 * @return int
 */
int main (void)
{
	int err;

    printk("\n*** Notification Device ***\n");

    /* Register SMP for DFU service */
#if CONFIG_MCUMGR_SMP_BT
    smp_bt_register();
#endif

	/* Initialization of distance measurement */
    err = app_dm_init();
    if (err) {
        return err;
    }

    /* Initialize BLE */
	err = bt_enable(NULL);
	if (err) {
		return err;
	}

    /* Load peripheral setting */
	err = settings_load();
	if (err) {
		return err;
	}

    /* Initialize NVS partition if fresh device. Else, read and update local DB */
    if (nvs_init()) {
        print_nvs_info();

        /* Set all the param to it's default */
        for (uint8_t idx = 0; idx < app_nvs.param.paired_device_count; idx++) {
            memcpy(paired_list[idx].id, app_nvs.param.pair_info[idx].id, BT_ADDR_SIZE);
            paired_list[idx].conn_status = CONN_STATUS_DISCONNECTED;
		    paired_list[idx].conn = NULL;
            paired_list[idx].pair_status = true;
            paired_list[idx].is_mobile = app_nvs.param.pair_info[idx].is_mobile;
        }

        dev_info.zone_range = app_nvs.param.zone_range;
    }

    /* Initialize all the necessary peripherals */
	timers_init();
    timers_context_init();
	buzzer_init();
	led_init();
	button_init();
    batt_init();

    /* Continue the operation based on device mode set */
	if (app_nvs.param.mode == DEVICE_MODE_DETECTION) {
		init_central();
	}

    /* Indicate to user, the device is ready */
    led_fast_blink(LED_COLOR_WHITE);
    NRFX_DELAY_US(500 * USEC_PER_MSEC);
    led_off();

	/* Define run state, default state: IDLE*/
    if (!app_nvs.param.paired_device_count) {
        app_update_state(APP_STATE_IDLE);
    }
    else {
        app_update_state(APP_STATE_PAIRING);
    }

    // Start the timer for battery monitoring
    //
    k_timer_start(&battery_monitor_timer, K_SECONDS(BATTERY_MONITOR_INTERVAL_SECS), K_NO_WAIT);

#if TEST_DFU
    app_update_state(APP_STATE_SETUP_FOR_OTA);
#endif

	return 0;
}
