/**
 * @file ns_dfu.c
 * @author Jithin A
 * @brief File handles all the DFU operations
 * @version 0.1
 * @date 2023-08-14
 *
 * @copyright Copyright (c) 2023
 *
 */

/* Includes --------------------------------------------------------------------------------------*/
#include "app_common.h"

// Zephyr
//
#include <zephyr/dfu/mcuboot.h>

// DFU
//
#if CONFIG_DFU_TARGET_MCUBOOT
#include <dfu/dfu_target.h>
#include <dfu/dfu_target_mcuboot.h>
#endif

#include "ns_dfu.h"

/* Private macros --------------------------------------------------------------------------------*/
// DFU param
// For DFU_TARGET_IMAGE_TYPE_MCUBOOT image type: given image pair index or -1 for all of image pair
// indexes. Disregard otherwise.
//
#define DFU_IMG_NUM                     (0)
#define DFU_STATUS_SUCCESS              (0)

/* Global param ----------------------------------------------------------------------------------*/
g_ota_ctx_t g_ota_ctx;

/* Private param ---------------------------------------------------------------------------------*/
// DFU stream buffer. Make sure the buffer is 4 bytes aligned.
//! NOTE: This buffer is handled by the DFU APIs, should not be used for any purposes
//
typedef union {
    uint32_t word[CONFIG_BT_BUF_ACL_RX_SIZE >> 2];
    uint8_t  bytes[CONFIG_BT_BUF_ACL_RX_SIZE];
} dfu_stream_buf_t;
static dfu_stream_buf_t g_dfu_stream;

/* Private function prototypes -------------------------------------------------------------------*/
static uint8_t bt_gen_crc_u8 (uint8_t *p_data, size_t length);
static void dfu_cb_abort (enum dfu_target_evt_id evt_id);
static int bt_ota_resume_req (uint8_t *p_ota_buf);
static int bt_ota_start_req (uint8_t *p_file_sz);
static int bt_ota_complete_req (void);
static int bt_ota_data_write (uint8_t *p_ota_buf);

/* _______________________________________________________________________________________________*/

/**
 * @brief bt_gen_crc_u8
 * Function generate the CRC_8 for the given bytes
 *
 * @param p_data   - data bytes
 * @param length   - size of the data to calculate CRC
 * @return uint8_t - return resultant CRC
 */
static uint8_t
bt_gen_crc_u8 (uint8_t *p_data, size_t length)
{
    uint8_t crc = 0x00;
    size_t i, j;

    for (i = 0; i < length; i++) {
        crc ^= p_data[i];
        for (j = 0; j < 8; j++) {
            if ((crc & 0x01) != 0) {
                crc = ((crc >> 1) ^ 0x8C);
            } else {
                crc >>= 1;
            }
        }
    }

    return crc;
}

/**
 * @brief dfu_cb_abort
 * Callback function if any error encountered during DFU - Abort DFU
 *
 */
static void
dfu_cb_abort (enum dfu_target_evt_id evt_id)
{
#if CONFIG_DFU_TARGET_MCUBOOT
    dfu_target_done(false);
#endif
}

/**
 * @brief bt_ota_resume_req
 * OTA resume request received, prepare DFU for OTA resume
 * We receive the last acknowledged or new OTA packet correlation ID, validate the received
 * correlation ID and send the respective response
 *
 * @param uint8_t *p_ota_buf
 * @return int
 */
static int
bt_ota_resume_req (uint8_t *p_ota_buf)
{
    int status = -1;

    // Read the battery status before OTA resume, proceed only if we have sufficient battery
    //
    uint8_t batt_perc = batt_read();
    if (batt_perc <= 20) {
        return status;
    }

#if 0 // CONFIG_DFU_TARGET_MCUBOOT
    uint16_t corr_id = 0;

    // Read the received correlation ID, skip packet header data
    //
    corr_id = swap_endian_pu8_to_u16(p_ota_buf + CMD_PKT_HEADER_SZ_BYTES);

    // Validate correlation ID, should be either last acknowledged or upcoming new ID
    //
    if ((corr_id == g_ota_ctx.corr_id) || (corr_id == g_ota_ctx.corr_id + 1)) {
        // Abort a download by calling the dfu_target_done() function, in which case the same
        // download can be resumed later by initializing the same target.
        //
        dfu_target_done(false);
        status = dfu_target_init(DFU_TARGET_IMAGE_TYPE_MCUBOOT, DFU_IMG_NUM, g_ota_ctx.file_sz,
                                 dfu_cb_abort);
    }
#endif
    return status;
}

/**
 * @brief bt_ota_start_req
 * OTA start request received, prepare for DFU
 *
 * @param p_msg
 * @return int
 */
static int
bt_ota_start_req (uint8_t *p_file_sz)
{
    int status = -1;

    // Read the battery status before start OTA, proceed only if we have sufficient battery
    //
    uint8_t batt_perc = batt_read();
    if (batt_perc <= 20) {
        return status;
    }

#if CONFIG_DFU_TARGET_MCUBOOT
    // Start fresh as this is the start request received
    //
    dfu_target_reset();

    // Setup stream buffer for DFU handling
    //
    status = dfu_target_mcuboot_set_buf(g_dfu_stream.bytes, sizeof(g_dfu_stream));
    if (status != DFU_STATUS_SUCCESS) {
        return status;
    }

    // Initialize the specific DFU target. Internally, it uses the image type and image number
    // to decide which target to initialize.
    // Use MCUboot-style firmware upgrades for updates to the application and to the upgradable
    // bootloader. This type of DFU writes the data given to the dfu_target_write() function
    // into the secondary slot specified by MCUboot’s flash memory partitions.
    // The file size will be received over BLE along command request
    //
    uint32_t file_sz = 0;
    memcpy(&file_sz, p_file_sz, sizeof(file_sz));
    status = dfu_target_init(DFU_TARGET_IMAGE_TYPE_MCUBOOT, DFU_IMG_NUM, file_sz, dfu_cb_abort);
#endif
    return status;
}

/**
 * @brief bt_ota_complete_req
 * OTA complete request received, process OTA
 *
 * @param p_msg
 * @return int
 */
static int
bt_ota_complete_req (void)
{
    int status = -1;

#if CONFIG_DFU_TARGET_MCUBOOT
    g_ota_ctx.complete = false;

    // Release the resources that were needed for the current DFU target.
    // Indicate whether the process completed successfully or was aborted.
    // If we are here it means the process completed successfully.
    //
    status = dfu_target_done(true);
    if (status != DFU_STATUS_SUCCESS) {
        return status;
    }

    status = dfu_target_schedule_update(DFU_IMG_NUM);
    if (status != DFU_STATUS_SUCCESS) {
        return status;
    }

    // Indicate OTA process is complete and boot the new image
    //
    g_ota_ctx.process  = false;
    g_ota_ctx.complete = true;
#endif

    return status;
}

/**
 * @brief bt_ota_data_write
 * Parse received OTA data and write the received image data into the secondary slot
 *
 * @param p_msg
 * @return int
 */
static int
bt_ota_data_write (uint8_t *p_ota_buf)
{
    int status = -1;

#if CONFIG_DFU_TARGET_MCUBOOT
    // Below are the param required for processing
    // Correlation id - used for acknowledging the received packet to the peer.
    // Data len       - actual OTA data len
    //

    uint16_t corr_id = swap_endian_pu8_to_u16(p_ota_buf);
    p_ota_buf += sizeof(corr_id);

    uint8_t data_len = *p_ota_buf++;
    uint8_t rcvd_crc = *p_ota_buf++;

    // Validate the received packet CRC
    //
    uint8_t crc = bt_gen_crc_u8(p_ota_buf, data_len);
    if (crc != rcvd_crc) {
        return status;
    }

    status = dfu_target_write((const void *)p_ota_buf, data_len);
    if (status == DFU_STATUS_SUCCESS) {
        g_ota_ctx.corr_id = corr_id;
    }
#endif

    return status;
}

/**
 * @brief bt_abort_dfu
 * Abort DFU process
 *
 */
void
bt_abort_dfu (void)
{
#if CONFIG_DFU_TARGET_MCUBOOT
    dfu_target_done(false);
#endif

    g_ota_ctx.process = false;
}

/**
 * @brief bt_process_ota
 * Process OTA data received from the TF-M as DFU is not supported in TF-M
 *
 * @param opcode - OTA operation to perform
 * @param p_data - pointer to the OTA buffer
 */
void
bt_process_ota (uint8_t *p_data)
{
#if CONFIG_DFU_TARGET_MCUBOOT
    int status = -1;
    uint8_t resp_data[4];
    uint8_t off = 0;
    resp_data[off++] = p_data[BYTE_OFF_CMD];

    if (p_data[BYTE_OFF_CMD] == OPCODE_OTA_DATA) {
        status = bt_ota_data_write(p_data);
        if (status != DFU_STATUS_SUCCESS) {
            return;
        }

        // Prepare response packet
        // This packet do not include any header data, should only contain response data
        //
        uint16_t corr_id   = swap_endian_pu8_to_u16((uint8_t *)&g_ota_ctx.corr_id);
        resp_data[off++] = sizeof(corr_id);
        memcpy(&resp_data[off], &corr_id, sizeof(corr_id));
    }
    else if (p_data[BYTE_OFF_CMD] == CMD_OTA_RESUME) {
        if (bt_ota_resume_req(p_data) == DFU_STATUS_SUCCESS) {
            g_ota_ctx.process = true;
        }

        // Prepare response packet
        //
        resp_data[off++] = sizeof(g_ota_ctx.process);
        resp_data[off] = g_ota_ctx.process;
    }
    else if (p_data[BYTE_OFF_CMD] == CMD_OTA_START) {
        if (bt_ota_start_req(&p_data[BYTE_OFF_DATA]) == DFU_STATUS_SUCCESS) {
            g_ota_ctx.process = true;
        }

        // Prepare response packet
        //
        resp_data[off++] = sizeof(g_ota_ctx.process);
        resp_data[off] = g_ota_ctx.process;
    }
    else if (p_data[BYTE_OFF_CMD] == CMD_OTA_COMPLETE) {
        // Indicate DFU system the OTA file download is complete
        //
        bt_ota_complete_req();

        // Prepare response packet
        //
        bt_pack_resp_data(CMD_OTA_COMPLETE, sizeof(g_ota_ctx.complete), &g_ota_ctx.complete,
                          sizeof(g_ota_ctx.complete), gu8_resp_buf, p_resp_ctx);
        bt_indicate(p_resp_ctx);
        k_msleep(500); // wait for the notification to reach to the peer.

        event_t event = EVENT_OTA_DOWNLOAD_COMPLETE;
        k_msgq_put(&main_thread_queue, &event, K_NO_WAIT);
    }

    // Send the response data over BLE
    //
    bt_indicate(resp_data, sizeof(resp_data));
#endif
}
